(* 

*)   

{
 open Parser
 exception Eof
 exception LexicalError of int
 exception NeverHappen
 
 fun verbose s = (* (print_string s; print_newline(); s) *) s 
 val outch = ref stdout
 type state = A | B 
 val nt = ref ","
 val tl = ref "."
 val rm = ref "/"
 val mcrdef = ref []
 val lid = ref A
 val stat = ref "normal"
 val nl = ref 1

 fun get_newsymbol s =
   let
     val first = String.index s '{'
     val second = String.index s '}'
     val last = String.sub s (second+1) ((String.length s)-(second+1))
     val third = String.index last '{'
     val forth = String.index last '}'
     val sym = String.sub s (first+1) (second - (first+1))
     val def = String.sub last (third+1) (forth - (third+1))
   in
     (sym,def)
   end 

 fun font_parse s tlist = 
   if s="" then (List.rev tlist)
     else (let val len = String.length s
               val first = String.sub s 0 1
               val ntidx = (String.index s (String.get !nt 0)) handle Not_found => len 
               val tlidx = (String.index s (String.get !tl 0)) handle Not_found => len
               val rmidx = (String.index s (String.get !rm 0)) handle Not_found => len
               fun min a b c = if a<b then (if a<c then a else c)
                                      else (if b<c then b else c)
               fun med a b c = if a<b then (if b<c then b else (if a<c then c else a))
                                      else (if a<c then a else (if b<c then c else b))
               val deli1 = min ntidx tlidx rmidx 
               val deli2 = med ntidx tlidx rmidx
          in
            if deli1=len 
              then font_parse (String.sub s deli1 (len-deli1)) ((s,"normal")::tlist) 
              else (if deli1=0 
                      then (if first=(!nt) then (stat:="nt"; font_parse (String.sub s deli2 (len-deli2)) (((String.sub s 1 (deli2-1)),(!stat))::tlist))
                              else (if first=(!tl) then (stat:="tl"; font_parse (String.sub s deli2 (len-deli2)) (((String.sub s 1 (deli2-1)),(!stat))::tlist))
                              else (if first=(!rm) then (stat:="rm"; font_parse (String.sub s deli2 (len-deli2)) (((String.sub s 1 (deli2-1)),(!stat))::tlist))
                              else raise NeverHappen )))
                      else (if (deli1=1)&&(first="\\") 
                              then font_parse (String.sub s deli2 (len-deli2)) (((String.sub s 1 (deli2-1)),(!stat))::tlist)
                              else font_parse (String.sub s deli1 (len-deli1)) (((String.sub s 0 deli1),"normal")::tlist)))
          end)

} 

val space = ' '
val tab = '\t'
val newline = '\n'
val bar = '-'('-')+
    (* KS X 1001 a.k.a. KS C 5601 or euc-kr *)
val hangul = (['\176'-'\200']['\161'-'\254'])+
    (* KS X 1005-1 a.k.a. KS C 5700, unicode, or ISO/IEC 10646-1 *)
val uhangul = (['\172'-'\214']['\000'-'\255']|'\215'['\000'-'\163'])+
val alphanum = ['a'-'z' 'A'-'Z']|hangul|uhangul|['0'-'9']
val reservedchar = ['^' '_' '"' '\\']
val reservedstr = "|-" | "|->" | "=>" | "<=" | "->" | "<-" | "(-" | "-)" | bar 
val escaped = "\\" reservedchar 
val unreserved = 
['<' '>' '&' '$' '{' '}' '%' '~' '!' '?' '#' '*' '+' '-' '/' '=' '|' '[' ']' ':' ';' '\'' ',' '.' '@' '`' '(' ')'] 

rule start = parse
    "\\" "begin" "{" "lantex" "}"
         { lid := B ; stat:="normal"; syntex_mode lexbuf}
  | eof { ignore (verbose "eof"); EOF}
  | newline {nl:=(!nl)+1; start lexbuf}
  | _ { start lexbuf}
  
and syntex_mode = parse
       "\\" "end" "{" "lantex" "}" {lid :=A; start lexbuf}
     | "\\" "ntfont" "{" unreserved "}"
         {let val s = (Lexing.lexeme lexbuf)
              val fontsym = String.sub s ((String.length s)-2) 1
          in 
            nt := fontsym
          end; stat:="normal"; syntex_mode lexbuf }
     | "\\" "tlfont" "{" unreserved "}"
         {let val s = (Lexing.lexeme lexbuf)
              val fontsym = String.sub s ((String.length s)-2) 1
          in 
            tl := fontsym
          end; stat:="normal"; syntex_mode lexbuf }     
     | "\\" "rmfont" "{" unreserved "}"
         {let val s = (Lexing.lexeme lexbuf)
              val fontsym = String.sub s ((String.length s)-2) 1
          in 
            rm := fontsym
          end; stat:="normal"; syntex_mode lexbuf }     
     | "\\" "newsymbol" "{" (alphanum|unreserved)+ "}" "{" ("\\"|alphanum|unreserved)+ "}"
         {let val s = (Lexing.lexeme lexbuf)
          in
            mcrdef := (get_newsymbol s)::(!mcrdef)
          end; stat:="normal"; syntex_mode lexbuf}
     | space {stat:="normal"; syntex_mode lexbuf}
     | tab {stat:="normal"; syntex_mode lexbuf }
     | newline {stat:="normal"; syntex_mode lexbuf}
     | "%Line###of###source" newline {stat:="normal"; nl:=(!nl)+1; syntex_mode lexbuf}
     | eof {raise Eof} 
     | "\\"? (alphanum | unreserved)+
         {let val s = (Lexing.lexeme lexbuf)
              val fst = String.sub s 0 1
              val last = (String.sub s 1 ((String.length s)-1)) handle _ => s
              val dfn = ref ""

              fun findsym str deflist =
                case deflist of
                  [] => false
                | (sym,def)::tl => if sym=str then (dfn:=def;true)
                                              else findsym str tl
          in
            if (findsym s !mcrdef) then MATHSYM (!dfn)
              else (if s = "|-" then VDASH
              else (if s = "|->" then MAPSTO
              else (if s = "=>" then RBIGARROW 
              else (if s = "<=" then LBIGARROW
              else (if s = "->" then RARROW
              else (if s = "<-" then LARROW
              else (if s = "(-" then IN
              else (if s = "-)" then NI
              else (if s = "<" then LANGLE
              else (if s = ">" then RANGLE
              else (if (Str.string_match (Str.regexp "--+") s 0) then BAR
              else (if fst="\\"
                      then (if (findsym last !mcrdef)
                            || (last = "|-")
                            || (last = "|->")
                            || (last = "=>")
                            || (last = "<=")
                            || (last = "->")
                            || (last = "<-")
                            || (last = "(-")
                            || (last = "-)")
                            || (last = "<")
                            || (last = ">")
                            || (Str.string_match (Str.regexp "--+") last 0) 
                              then VAR [(last,"normal")]
                              else VAR (font_parse s []))
              else VAR (font_parse s [])
              ))))))))))))
          end} 
     | "^" {HAT(!nl)} 
     | "_" {UNDERBAR(!nl)}
     | "\"" {DQ(!nl)}
     | "\\" {ESCAPE (!stat)} 
     | _ {raise LexicalError(!nl)} 

and rm_syntex = parse
    "\\" "begin" "{" "lantex" "}" 
         { rm_syntex lexbuf}
  | "\\" "end" "{" "lantex" "}"
         { rm_syntex lexbuf}  
  | "\\" "ntfont" "{" unreserved "}"
         { rm_syntex lexbuf}  
  | "\\" "tlfont" "{" unreserved "}"
         { rm_syntex lexbuf}  
  | "\\" "rmfont" "{" unreserved "}"
         { rm_syntex lexbuf}  
  | "\\" "newsymbol" "{" (alphanum|unreserved)+ "}" "{" ("\\"|alphanum|unreserved)+ "}"
         { rm_syntex lexbuf}  
  | "%Line###of###source" newline {rm_syntex lexbuf}
  | eof { ignore (verbose "eof"); EOF}
  | _ {output_string !outch (Lexing.lexeme lexbuf); rm_syntex lexbuf}
{

  fun my_lex lexbuf  = 
     if (!lid = A) then start lexbuf else syntex_mode lexbuf
   
}   

