exception WrongFileFormat of string


fun remove_ext file ex =
  let 
    val len = String.length file
    val dot = (String.rindex file '.') handle _ => raise WrongFileFormat(file)
    val ext = (String.sub file dot (len-dot))
  in
    if(ext=((".")^ex)) then String.sub file 0 dot
                 else raise WrongFileFormat(file)
  end
    
fun remove_tag fid =
  let
     val ic = open_in fid 
     val outf = String.concat (remove_ext fid "mid") ["",".tex"]
     val oc = open_out outf
     val _ = (Lexer.outch := oc)
     val lexbuf = Lexing.from_channel ic 
     val tok = Lexer.rm_syntex lexbuf
  in
    close_in ic;close_out oc
  end

fun trans_syntex fid =
  let
     val ic = open_in fid 
     val outf = String.concat (remove_ext fid "tab") ["",".mid"]
     val oc = open_out outf
     val lexbuf = Lexing.from_channel ic 
     val pgm = Parser.program Lexer.my_lex lexbuf
  in
     close_in ic;
     Ast.runtop oc pgm fid;
     close_out oc;
     print_newline();
     remove_tag outf;
     print_newline()
  end
    
fun run fid =
  let
     val outf = String.concat (remove_ext fid "l") ["",".tab"]
  in
     print_newline();
     print_newline();
     Array.make_array fid;
     print_newline();
     trans_syntex outf
  end

val _ = run (Sys.argv.[1]) handle Lexer.Eof => exit 0 
                                | WrongFileFormat (f) => print_string "The file ";print_string f; print_string " is not a LanTeX file.\n"
                                | Array.Invalid_open(n) => print_string "ERROR(Line: "; print_int n; print_string ") - An invalid paragraph opening\n"
                                | Array.Invalid_close(n) => print_string "error(line: "; print_int n; print_string ") - An invalid paragraph closing\n"
                                | Lexer.LexicalError(n) => print_string "error(line: "; print_int n; print_string ") - Lexical Error\n"
                                | Parser.ParseError(n,s) => print_string "error(line: "; print_int n; print_string ") - Wrong usage with ";print_string s;print_string "\n"
