
%{
open Format
open Parsing

exception ParseError of int*string

fun eloc n = (rhs_start n,rhs_end n)
fun sloc () = (symbol_start(), symbol_end())
fun vb s = print_string s; print_newline (); ()
%}
  
%token EOF PERCENT 
%token VDASH MAPSTO IN NI BAR
%token RBIGARROW LBIGARROW RARROW LARROW LANGLE RANGLE

%token <Ast.sslist> VAR
%token <string> MATHSYM
%token <string> ESCAPE

%token <string> NTSTR
%token <string> TLSTR
%token <string> CMSTR

%token <string> NSYMBOL

%token <int> HAT
%token <int> UNDERBAR
%token <int> DQ

%start program 
%start getfont
%start macrodef 
%type <Ast.top> program
%type <Font.fontall> getfont
%type <Macro.macrotop> macrodef
%%

////////// get font sign ///////////

getfont : fontstrs EOF   {Font.Top($1)}
        ;
        
fontstrs : fonts        {[$1]}
         | fonts fontstrs {$1::$2}
         ;

fonts : NTSTR   {Font.NtStr($1)}
      | TLSTR   {Font.TlStr($1)}
      | CMSTR   {Font.CmStr($1)}
      ;

/////////// find macro definition ////////////////

macrodef : macrostr EOF {Macro.Top($1)}
         ;

macrostr : nsymbol         {[$1]}
         | nsymbol macrostr {$1::$2}
         ;

nsymbol : NSYMBOL       {Macro.NSymbol($1)}
        ;

////////// LanTeX to LaTeX //////////

program : toplevels EOF       {Ast.Top($1)}
        ;

toplevels : toplevel             {[$1]}
          | toplevel toplevels   {$1::$2}
          ;

toplevel : syntex             {Ast.Syntex($1,sloc())}
         | latex            {Ast.Latex($1,sloc())}
        ;

latex : ESCAPE  {Ast.LatexEscChar($1)}
      | escaped         {Ast.LatexEsc($1)}
      ;

syntex : exp                    {Ast.Exp($1)}
       | rule                   {Ast.Rule($1)}
       ;

exp : primary   {Ast.Prim($1)}
    | updown    {Ast.Updn($1)}
    ;

primary : VAR           {Ast.Var($1)}
        | reservedstr           {Ast.PrRstr($1)}
        | angle                 {Ast.PrAng($1)}
        | MATHSYM       {Ast.Mathsym($1)}
        ;
 
updown : HAT primary            {Ast.SupPr($2)}
       | HAT latex              {Ast.SupLa($2)}
       | HAT error              {raise ParseError($1,"^")}
       | UNDERBAR primary       {Ast.SubPr($2)}
       | UNDERBAR latex         {Ast.SubLa($2)}
       | UNDERBAR error         {raise ParseError($1,"_")}
       ;


content : exp              {Ast.RuleCont_Exp($1)}
        | latex            {Ast.RuleCont_Latex($1)}
        ;
        
contents : content      {[$1]}
         | content contents {$1::$2}
     ;

rule : DQ DQ BAR DQ contents DQ        {Ast.RuleAxiom($5)}
     | DQ contents DQ BAR DQ contents DQ    {Ast.RuleBasic($2,$6)}
     | rule BAR DQ contents DQ    {Ast.RuleNest($1,$4)}
     | DQ error                         {raise ParseError($1,"\"")}
     ;

reservedchar : HAT      {Ast.Hat}
             | UNDERBAR {Ast.Underbar}
             | DQ       {Ast.Dq}
             ;
             
reservedstr : VDASH           {Ast.Vdash}
            | MAPSTO          {Ast.Mapsto}
            | RBIGARROW       {Ast.Rbigarrow }
            | LBIGARROW       {Ast.Lbigarrow }
            | RARROW          {Ast.Rarrow }
            | LARROW          {Ast.Larrow }
            | IN              {Ast.In }
            | NI              {Ast.Ni }
            ;
            
angle : LANGLE   {Ast.Langle}
      | RANGLE   {Ast.Rangle}
             ;

escaped : ESCAPE reservedchar   {Ast.Escaped($1,$2)}
        ;

%%
