structure Array =
struct

exception Exit_loop
exception WrongFileFormat of string
exception Invalid_open of int
exception Invalid_close of int
exception Stop1
exception Stop2


fun remove_ext file =
  let 
    val len = String.length file
    val dot = (String.rindex file '.') handle _ => raise WrongFileFormat(file)
    val ext = (String.sub file dot (len-dot))
  in
    if(ext=".l") then String.sub file 0 dot
                 else raise WrongFileFormat(file)
  end

fun arr_st str l = 
  let val len = String.length str
      val bslash = (String.index str '\\')handle Not_found => (len+1)
      val bracket = (String.index str '[')handle Not_found => (len+1)
      val esc = (String.index str '`')handle Not_found => (len+1)
      fun small a b = if(a<b)then a else b
      val first = small bslash bracket
  in
    if (bslash = (len+1)) then (false,l+len)
      else (if (bracket=(len+1)) then (false,l+len)
        else (if (not(bracket=(bslash+1))) then arr_st (String.sub str (first+1) (len-first-1)) (l+first+1) 
          else (if (bslash=(esc+1)) then arr_st (String.sub str (bslash+1) (len-bslash-1)) (l+bslash+1)
            else (true,bslash+l)))) 
  end

fun arr_end str l = 
  let val len = String.length str
      val bslash = (String.index str '\\')handle Not_found => (len+1)
      val bracket = (String.index str ']')handle Not_found => (len+1)
      val esc = (String.index str '`')handle Not_found => (len+1)
      fun small a b = if(a<b)then a else b
      val first = small bslash bracket
  in
    if (bslash = (len+1)) then (false,l+len)
      else (if (bracket=(len+1)) then (false,l+len)
        else (if (not(bracket=(bslash+1))) then arr_end (String.sub str (first+1) (len-first-1)) (l+first+1) 
          else (if (bslash=(esc+1)) then arr_st (String.sub str (bslash+1) (len-bslash-1)) (l+bslash+1)
            else (true,bslash+l)))) 
  end

fun begin_lantex s =
  if (not(s=""))&&((String.sub s 0 1)="\\")
    then (Str.string_match (Str.regexp "begin{lantex}") s 1) else false

fun end_lantex s =
  if (not(s=""))&&((String.sub s 0 1)="\\")
    then (Str.string_match (Str.regexp "end{lantex}") s 1) else false
 

fun get_token str tokenlist =
  if (str="")
    then tokenlist
    else (
      let val len = String.length str
          val sp = (String.index str ' ') handle Not_found => (len)
          val tab = (String.index str '\t') handle Not_found => (len)
          fun small a b = if (a<b) then a else b
          val index = small sp tab
          val token = (if (index=0)
                         then String.sub str 0 1
                         else String.sub str 0 index)
          val substr = (if (index=0)
                         then String.sub str 1 (len-1)
                         else String.sub str (index) (len-index))
      in 
        get_token substr (token::tokenlist)
      end)

fun loc2loc inlist outlist =
  case inlist of
    [] => outlist
  | (s,(l1,l2))::tl => loc2loc tl ((s,(l1,(l1+l2-1)))::outlist) 

fun joinsp inlist outlist sp loc =
  case inlist of
    [] => if (sp=0) then outlist
                    else (" ",(loc-sp,sp))::outlist
  | hd::tl => case hd of
                (" ",(l1,l2)) => joinsp tl outlist (sp+l2) (l1+l2)
              | ("\t",(l1,l2)) => joinsp tl outlist (sp+l2) (l1+l2)
              | (_,(l1,l2)) => if (not(sp=0))
                                 then joinsp tl (hd::((" ",(l1-sp,sp))::outlist)) 0 (l1+l2)
                                 else joinsp tl (hd::outlist) 0 (l1+l2)
    

fun findloc inlist outlist loc =
  case inlist of
    [] => outlist
  | hd::tl => case hd of
                " " => findloc tl ((" ",(loc,1))::outlist) (loc+1)
              | "\t" => findloc tl (("\t",(loc,(8-(loc%8))))::outlist) (loc+(8-(loc%8)))
              | _ => findloc tl ((hd,(loc,(String.length hd)))::outlist) (loc+(String.length hd))

fun match (x1,y1) (x2,y2) =
  if (x1<x2) 
    then (if (y1>y2) 
           then (true)
           else (if (x2<=y1) then (true) else false))
    else (if (y1<y2)
           then (true)
           else (if (x1<=y2) then (true) else false))

fun print_array oc col lst =
  case lst of 
    []=>(output_string oc " \\\\\n"; output_string oc "%Line###of###source\n") 
  | (s,(l1,l2))::tl => 
      (case s of 
        " " => let val (s1,s2) = (List.hd col)handle _ => (l2+1,l2+1)
                   val last = (if col=[] then [] else List.tl col)
                   val (s3,s4) = (List.hd last)handle _ => (l2+1,l2+1)
               in if (match (l1,l2) (s1,s2)) 
                    then ((if (s2-s1)=0 then output_string oc " && "
                            else (output_string oc " &\\hspace{";
                                  output_string oc (string_of_int (s2-s1+1));
                                  output_string oc "mm}& "));
                          if (match (l1,l2) (s3,s4))
                            then print_array oc last lst
                            else print_array oc last tl)
                    else (for i=0;i<=(l2-l1);i+1 do output_string oc "\\; " end;
                          print_array oc col tl)
               end
      | _ => output_string oc s; print_array oc col tl)


fun sort inlist =
  let fun (>>>) (x1,y1) (x2,y2) = if x1<x2 then -1 else 1
  in List.sort (>>>) inlist
  end

fun find_sp inlist outlist =
  case inlist of
    [] => List.rev outlist 
  | hd::tl => case hd of
                (" ",(s,e)) => find_sp tl ((s,e)::outlist)
              | _ => find_sp tl outlist

fun find_col ta spcol =
  case ta of
    [] => sort spcol
  | hd::tl =>(let val splist = ref (find_sp hd [])
                  val new = ref []
                  val slist = ref spcol
              in
                (while (not(!splist=[])) do 
                  (let val (x1,y1) = List.hd (!splist)
                  in    
                    (while (not(!slist=[])) do
                      (let val (x2,y2) = List.hd (!slist)
                      in
                        if (x1<x2) then(if (y1>y2) then new:=((x2,y2)::(!new))
                                                   else (if (y1>=x2) then new:=((x2,y1)::(!new))
                                                                    ))
                                   else (if(y1<y2) then new:=((x1,y1)::(!new))
                                                   else (if(x1<=y2) then new:=((x1,y2)::(!new))
                                                                   ))  
                      end);
                      slist := List.tl (!slist)
                    end)
                  end);
                  splist := List.tl (!splist);
                  slist := spcol
                end);
                find_col tl (!new)
              end)

fun find_list str =
  let val tlist1 = List.rev (get_token str [])
      val tlist2 = List.rev (findloc tlist1 [] 1) 
      val llist = List.rev (loc2loc (List.rev (joinsp tlist2 [] 0 1)) [])
  in llist
  end

fun print_plist oc lst =
  case lst of
    [] => output_string oc "\n"; output_string oc "%Line###of###source\n"
  | (s,(l1,l2))::tl => (case s of
                          " " => for i=0;i<=(l2-l1);i+1 do output_string oc " " end
                        | _ => output_string oc s);
                       print_plist oc tl

fun print_plain oc arr =
  case arr of
    [] => ()
  | hd::tl => print_plist oc hd; print_plain oc tl

fun blank_line str =
  case str of
    "" => true
  | _ => if (((String.sub str 0 1)=" ")||((String.sub str 0 1)="\t")) 
           then (if ((String.length str)>1) 
                  then blank_line (String.sub str 1 ((String.length str)-1))
                  else blank_line "")
           else false 

fun combine inlist outlist col (c1,c2) dummy =
  case inlist of
    [] => List.rev ((c1,c2)::outlist)
  | (s,(l1,l2))::tl => 
      case s of
        " " => (let val (s1,s2) = (List.hd col) handle _ => (l2+1,l2+1)
                in
                  if (l2<s1) 
                    then (if (col=[]) && dummy 
                           then combine tl outlist col (0,0) false
                           else (if l1=1 then combine tl outlist col (l2+1,l2+1) false 
                                         else combine tl outlist col (c1,c2) false))
                    else (if (l1<=s1)&&(s2<=l2)
                           then (if dummy then (if s2=l2 
                                                 then combine tl ((0,0)::outlist) (List.tl col) (s2+1,s2+1) true
                                                 else combine ((" ",(s2,l2))::tl) ((0,0)::outlist) (List.tl col) (l2+1,l2+1) true)
                                          else (if s2=l2
                                                 then combine tl ((c1,c2)::outlist) (List.tl col) (s2+1,s2+1) true
                                                 else combine ((" ",(s2,l2))::tl) ((c1,c2)::outlist) (List.tl col) (l2+1,l2+1) true))
                           else raise Stop1)
                end)
      | x => (let val (s1,s2) = (List.hd col) handle _ => (l2+1,l2+1)
              in
                if (l2<s1) 
                  then combine tl outlist col (c1,l2) false
                  else raise Stop2
              end)

fun put_col inlist col outlist loc =
  case inlist of
    [] => List.rev outlist
  | (l1,l2)::tl => let val (s1,s2) = List.hd col
                   in 
                     if loc=1 then put_col tl col ((l1,l2)::outlist) (loc+1)
                              else put_col tl (List.tl col) ((l1,l2)::((s1,s2)::outlist)) (loc+1)
                   end

fun find_align arr col alist =
  case arr of
    [] => List.rev alist
  | hd::tl => find_align tl col ((put_col (combine hd [] col (1,1) true ) col [] 1)::alist)

fun align arr col  =
  let val slist = find_align arr col []
      val outer = ref slist
      val inner = ref (List.hd slist)
      val outernum = List.length (!outer)
      val innernum = List.length (!inner)
      val flatlist = List.flatten slist
      val k = ref 0
      val result = ref []
      val first = ref 0
      val last = ref 0
      val left = ref true
      val right = ref true
  in 
    (for i=0;i<innernum;i+1 do 
      k:=i;
      (let val (a,b) = (List.nth flatlist !k)
      in first:=a; last:=b
      end);
      (for j=0;j<outernum;j+1 do
        (let val (f,l) = (List.nth flatlist !k)
        in
          (if (((!first)=0)&&((!last=0))) then (first:=f;last:=l));
          if ((not(f=0))&&(not(l=0))) then(
            left:= (f=(!first))&&(!left);
            right:= (l=(!last))&&(!right);
            first:=f; last:=l)
        end);
        k:= !k+innernum
      end);
      (if (!left) then (if (!right) then result:="c"::(!result)
                                   else result:="l"::(!result))
                 else (if (!right) then result:="r"::(!result)
                                   else result:="c"::(!result)));
      left:=true; right:=true
    end);
    List.rev !result
  end
     
fun print_align oc lst =
  case lst of
    [] => ()
  | hd::tl => output_string oc hd; print_align oc tl

fun find_fontdef s =
  (Str.string_match (Str.regexp "ntfont") s 0) 
  || (Str.string_match (Str.regexp "tlfont") s 0) 
  || (Str.string_match (Str.regexp "rmfont") s 0) 


fun fill_each_line inlist rtnlist max =
  case inlist of
    [] => List.rev rtnlist 
  | (s,(l1,l2))::tl => if tl=[] then (if l2<max then (if s=" " then fill_each_line tl ((" ",(l1,max))::rtnlist) max
                                                               else fill_each_line tl ((" ",((l2+1),max))::((s,(l1,l2))::rtnlist)) max)
                                                else fill_each_line tl ((s,(l1,l2))::rtnlist) max)
                                else fill_each_line tl ((s,(l1,l2))::rtnlist) max

fun fill_line ta rtnta max =
  case ta of
    [] => List.rev rtnta 
  | hd::tl => fill_line tl ((fill_each_line hd [] max)::rtnta) max 

fun make_array fid =
  let
    val ic = open_in fid 
    val ln = ref 0
    val oc = open_out (String.concat (remove_ext fid) ["",".tab"])
    val arr = ref false
    val tokenarray = ref []
    val line = ref 0
    val lantex = ref false
    val max = ref 0
  in
    (while true do
      (let val instr = input_line ic
           val _ = (ln:=(!ln)+1)
           val s = if (blank_line instr) then ref "" else ref instr
           val len = String.length !s
           val _ = if (begin_lantex instr) then (lantex:=true) 
                   else (if (end_lantex instr) then (lantex:=false) else lantex:=(!lantex))
       in
         if (not(!lantex)) then (output_string oc instr; output_string oc "\n") else (
         if (!s="") then (if (!arr) then s:=" ");
         (if (!max)<len then max:=len);
         while (not(!s="")) do
          if (!arr) 
           then(
             let val (st,ls) = arr_st !s 0
                 val (ed,le) = arr_end !s 0
             in
               if ed 
                 then (if st
                   then (if (le<ls) 
                           then (tokenarray := List.rev (!tokenarray);
                                 tokenarray := fill_line (!tokenarray) [] (!max);
                                (let val col = (find_col !tokenarray (find_sp (List.hd !tokenarray) [])) handle _ => []
                                     val cnum = List.length col
                                     val aline = String.sub !s 0 le
                                 in
                                   if (!line<=1||cnum=0) then(
                                       output_string oc "\n\\[\n";
                                       output_string oc "%Line###of###source\n";
                                       print_plain oc !tokenarray;
                                       output_string oc aline;
                                       output_string oc "\\]\n";
                                       output_string oc "%Line###of###source\n"
                                     ) else(
                                       output_string oc "\n\\[\n";
                                       output_string oc "%Line###of###source\n";
                                       output_string oc "\\begin{array}{";
                                       print_align oc (align !tokenarray col);
                                       output_string oc "}\n";
                                       List.iter (print_array oc col) (!tokenarray);
                                       output_string oc "\n\\end{array}";
                                       output_string oc "\n\\]\n";
                                       output_string oc "%Line###of###source\n")
                                 end);
                                 tokenarray := [];
                                 line := 0;
                                 max := 0;
                                 arr:=false;
                                 s := (String.sub !s (le+2) ((String.length !s)-(le+2)))
                           )else raise Invalid_open(!ln)
                   )else (tokenarray := List.rev (!tokenarray);   
                          tokenarray := fill_line (!tokenarray) [] (!max); 
                          (let val col = (find_col !tokenarray (find_sp (List.hd !tokenarray) [])) handle _ => []
                              val cnum = List.length col
                              val aline = String.sub !s 0 le
                          in
                            if (!line<=1||cnum=0) then (
                                output_string oc "\n\\[\n";
                                output_string oc "%Line###of###source\n";
                                print_plain oc !tokenarray;
                                output_string oc aline;
                                output_string oc "\\]\n";
                                output_string oc "%Line###of###source\n"
                              )else (
                                output_string oc "\n\\[\n";
                                output_string oc "%Line###of###source\n";
                                output_string oc "\\begin{array}{";
                                print_align oc (align !tokenarray col);
                                output_string oc "}\n";
                                List.iter (print_array oc col) (!tokenarray);
                                output_string oc "\n\\end{array}";
                                output_string oc "\n\\]\n";
                                output_string oc "%Line###of###source\n")
                          end);
                          tokenarray := [];
                          line := 0;
                          max := 0;
                          arr:=false;
                          s := (String.sub !s (le+2) ((String.length !s)-(le+2))))
                 )else (if st 
                          then raise Invalid_open(!ln)
                          else (tokenarray := ((find_list (!s))::(!tokenarray)) ; s:=""; line:=(!line)+1))
               end)
           else(
             let val (st,ls) = arr_st !s 0
                 val (ed,le) = arr_end !s 0
             in
               if st
                  then (if ed
                          then (if (ls<le) 
                                  then (output_string oc (String.sub !s 0 ls);
                                        arr:=true;
                                        s:= (String.sub !s (ls+2) ((String.length !s)-(ls+2)))
                                  )else raise Invalid_close(!ln)
                          )else (output_string oc (String.sub !s 0 ls);
                                 arr:=true;
                                 s := (String.sub !s (ls+2) ((String.length !s)-(ls+2))))
                  )else (if ed
                           then raise Invalid_close(!ln)
                           else output_string oc !s;output_string oc "\n";output_string oc "%Line###of###source\n"; s:="")
             end)
         end)
       end) handle End_of_file => raise Exit_loop
       
     end) handle Exit_loop => 
    close_in ic;
    close_out oc
  end
end
